package FL;

import java.io.*;
import static FL.Scanner.TokenType;
import static FL.Scanner.TokenType.*;

public class Parser {

	protected Scanner scanner;

	public Parser(Scanner scanner) {
		this.scanner=scanner;
	}

	static boolean isBinaryOp(TokenType op) {
		switch(op){
		     case PLUS:
		     case TIMES:
		     case OR:
		     case AND:
		     case EQUAL:
			     return true;
		}
		return false;
	}

	public Prog parseProg(boolean interactiveMode) throws ParseException,IOException {
		scanner.nextToken();
		Prog p=new ExpProg(parseExp());
		if (interactiveMode) {
			if (scanner.ttype!=SEMICOLON)
				throw new ParseException("programs must be ended by ';'");
		} else	if (scanner.ttype!=EOF)
				throw new ParseException("unexpected token at the end of the program");
		return p;   
	}

	void expecting(TokenType tok) throws ParseException, IOException {
		expecting(tok, "expecting "+tok);
	}

	void expecting(TokenType tok, String message) throws ParseException, IOException {
		if (scanner.ttype==tok)
			scanner.nextToken();
		else	throw new ParseException(message);
	}

	Exp parseBoolean() throws IOException {
		TokenType currentTT = scanner.ttype;
		scanner.nextToken();
		return currentTT==TRUE ? BoolExp.TRUE : BoolExp.FALSE;
	}

	Exp parseExp() throws ParseException,IOException {
		switch (scanner.ttype){
		case OROUND: return parseRound();
		case NUM: return parseNum();
		case IF: return parseIf();
		case LET: return parseLet();
		case ID: return parseId();
		case TRUE: 
		case FALSE: return parseBoolean();
		}
		throw new ParseException("syntax error while reading an expression");
	}

	Exp parseRound() throws ParseException,IOException {
		scanner.nextToken();
		Exp exp1=parseExp();
		TokenType op=scanner.ttype;
		if (!isBinaryOp(op))
			throw new ParseException("expecting a binary operator");
		scanner.nextToken();
		Exp exp2=parseExp();
		expecting(CROUND, "expecting ')'");
		switch(op) {
		case PLUS: return new SumExp(exp1,exp2);
		case TIMES: return new ProdExp(exp1,exp2);
		case OR: return new OrExp(exp1,exp2);
		case AND: return new AndExp(exp1,exp2);
		case EQUAL: return new EqExp(exp1,exp2);
		}
		throw new ParseException("Internal error in parseExp(); op="+op);
	}

	NumExp parseNum() throws ParseException,IOException {
		int nval=scanner.nval;
		scanner.nextToken();
		return new NumExp(nval);
	}

	IfExp parseIf() throws ParseException,IOException {
		scanner.nextToken();
		Exp ifExp=parseExp();
		expecting(THEN);
		Exp thenExp=parseExp();
		expecting(ELSE);
		Exp elseExp=parseExp();
		expecting(ENDIF);
		return new IfExp(ifExp,thenExp,elseExp);
	}

	LetExp parseLet() throws ParseException,IOException {
		scanner.nextToken();
		Decs decs=parseDecs();
		expecting(IN);
		Exp exp=parseExp();
		expecting(ENDLET);
		return new LetExp(decs,exp);
	}

	Exp parseId() throws ParseException,IOException {
		Id id=new Id(scanner.sval);
		if (scanner.nextToken()!=OROUND) 
			return new IdExp(id);
		scanner.nextToken();
		Exps exps=parseExps();
		expecting(CROUND, "expecting ')'");
		return new FunCallExp(id,exps);       
	}

	Decs parseDecs() throws ParseException,IOException {
		DecsImpl decs=new DecsImpl(parseDec());
		while (scanner.ttype==SEMICOLON) {
			scanner.nextToken();
			decs.add(parseDec());
		}
		return decs;
	}

	Dec parseDec() throws ParseException,IOException {
		if (scanner.ttype!=ID)
			throw new ParseException("expecting an identifier");
		Id id=new Id(scanner.sval);
		switch(scanner.nextToken()) {
		case COLON: return parseConstDec(id);
		case OROUND: return parseFunDec(id);
		}
		throw new ParseException("expecting either ':' or '('");
	}

	ConstDec parseConstDec(Id id) throws ParseException,IOException {
		scanner.nextToken();
		PrimType type=parsePrimType();
		expecting(EQUAL, "expecting '='");
		Exp exp=parseExp();
		return new ConstDec(id,type,exp);
	}

	FunDec parseFunDec(Id id) throws ParseException,IOException {
		scanner.nextToken();
		Params params=parseParams();
		expecting(CROUND, "expecting ')'");                       
		expecting(COLON, "expecting ':'");                       
		PrimType retType=parsePrimType();
		expecting(EQUAL, "expecting '='");
		Exp exp=parseExp();
		return new FunDec(id,params,retType,exp);
	}

	Params parseParams() throws ParseException,IOException { 
		ParamsImpl params=new ParamsImpl(parseParam());
		while (scanner.ttype==COMMA) {
			scanner.nextToken();
			params.add(parseParam());
		}
		return params;
	}

	Param parseParam() throws ParseException,IOException { 
		if (scanner.ttype!=ID)
			throw new ParseException("expecting an identifier");  
		Id id=new Id(scanner.sval);
		scanner.nextToken();
		expecting(COLON, "expecting ':'");
		return new PrimTypeParam(id, parsePrimType());
	} 

	PrimType parsePrimType() throws ParseException,IOException {
		switch(scanner.ttype){
		case INT:
			scanner.nextToken();
			return PrimType.INT;
		case BOOL:
			scanner.nextToken();
			return PrimType.BOOL;
		}
		throw new ParseException("expecting a primitive type");
	}

	Exps parseExps() throws ParseException,IOException {
		ExpsImpl exps=new ExpsImpl(parseExp());
		while (scanner.ttype==COMMA){
			scanner.nextToken();
			exps.add(parseExp());
		}
		return exps;
	}

} 
